/* sds_stddev.c: SDS 940 standard devices

   Copyright (c) 2001, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   ptr		paper tape reader
   ptp		paper tape punch
   tti		keyboard
   tto		teleprinter
*/

#include "sds_defs.h"
#define TT_CR		052				/* typewriter */
#define TT_TB		072
#define TT_BS		032

extern uint32 xfr_req;
extern int32 stop_invins, stop_invdev, stop_inviop;
int32 ptr_ch, ptr_ac;					/* curr, asgnd ch */
int32 ptr_sor = 0;					/* start of rec */
int32 ptr_stopioe = 1;					/* stop on err */
int32 ptp_ch, ptp_ac;
int32 ptp_ldr = 0;					/* no leader */
int32 ptp_stopioe = 1;
int32 tti_ch, tti_ac;
int32 tto_ch, tto_ac;
t_stat ptr_svc (UNIT *uptr);
t_stat ptr_reset (DEVICE *dptr);
void ptr_set_err (void);
t_stat ptp_svc (UNIT *uptr);
t_stat ptp_reset (DEVICE *dptr);
t_stat ptp_out (int32 dat);
void ptp_set_err (void);
t_stat tti_svc (UNIT *uptr);
t_stat tti_reset (DEVICE *dptr);
t_stat tto_svc (UNIT *uptr);
t_stat tto_reset (DEVICE *dptr);

extern const char ascii_to_sds[128];
extern const char sds_to_ascii[64];
extern const char odd_par[64];

/* PTR data structures

   ptr_dev	PTR device descriptor
   ptr_unit	PTR unit
   ptr_reg	PTR register list
*/

UNIT ptr_unit = {
	UDATA (&ptr_svc, UNIT_SEQ+UNIT_ATTABLE+UNIT_ROABLE, 0),
		SERIAL_IN_WAIT };

REG ptr_reg[] = {
	{ ORDATA (BUF, ptr_unit.buf, 7) },
	{ FLDATA (XFR, xfr_req, XFR_V_PTR) },
	{ FLDATA (SOR, ptr_sor, 0) },
	{ ORDATA (CHAN, ptr_ch, 4) },
	{ DRDATA (POS, ptr_unit.pos, 31), PV_LEFT },
	{ DRDATA (TIME, ptr_unit.wait, 24), REG_NZ + PV_LEFT },
	{ FLDATA (STOP_IOE, ptr_stopioe, 0) },
	{ ORDATA (ASCH, ptr_ac, 4), REG_HRO },
	{ NULL }  };

MTAB ptr_mod[] = {
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL",
		"CHANNEL", &set_chan, &show_chan, &ptr_ac },
	{ 0 }  };


DEVICE ptr_dev = {
	"PTR", &ptr_unit, ptr_reg, ptr_mod,
	1, 10, 31, 1, 8, 8,
	NULL, NULL, &ptr_reset,
	NULL, NULL, NULL };

/* PTP data structures

   ptp_dev	PTP device descriptor
   ptp_unit	PTP unit
   ptp_reg	PTP register list
*/

UNIT ptp_unit = {
	UDATA (&ptp_svc, UNIT_SEQ+UNIT_ATTABLE, 0), SERIAL_OUT_WAIT };

REG ptp_reg[] = {
	{ ORDATA (BUF, ptp_unit.buf, 7) },
	{ FLDATA (XFR, xfr_req, XFR_V_PTP) },
	{ FLDATA (LDR, ptp_ldr, 0) },
	{ ORDATA (CHAN, ptp_ch, 4) },
	{ DRDATA (POS, ptp_unit.pos, 31), PV_LEFT },
	{ DRDATA (TIME, ptp_unit.wait, 24), REG_NZ + PV_LEFT },
	{ FLDATA (STOP_IOE, ptp_stopioe, 0) },
	{ ORDATA (ASCH, ptp_ac, 4), REG_HRO },
	{ NULL }  };

MTAB ptp_mod[] = {
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL", "CHANNEL",
		&set_chan, &show_chan, &ptp_ac },
	{ 0 }  };

DEVICE ptp_dev = {
	"PTP", &ptp_unit, ptp_reg, ptp_mod,
	1, 10, 31, 1, 8, 8,
	NULL, NULL, &ptp_reset,
	NULL, NULL, NULL };

/* TTI data structures

   tti_dev	TTI device descriptor
   tti_unit	TTI unit
   tti_reg	TTI register list
*/

UNIT tti_unit = { UDATA (&tti_svc, 0, 0), KBD_POLL_WAIT };

REG tti_reg[] = {
	{ ORDATA (BUF, tti_unit.buf, 6) },
	{ FLDATA (XFR, xfr_req, XFR_V_TTI) },
	{ ORDATA (CHAN, tti_ch, 4) },
	{ DRDATA (POS, tti_unit.pos, 31), PV_LEFT },
	{ DRDATA (TIME, tti_unit.wait, 24), REG_NZ + PV_LEFT },
	{ ORDATA (ASCH, tti_ac, 4), REG_HRO },
	{ NULL }  };

MTAB tti_mod[] = {
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL", "CHANNEL",
		&set_chan, &show_chan, &tti_ac },
	{ 0 }  };

DEVICE tti_dev = {
	"TTI", &tti_unit, tti_reg, tti_mod,
	1, 10, 31, 1, 8, 8,
	NULL, NULL, &tti_reset,
	NULL, NULL, NULL };

/* TTO data structures

   tto_dev	TTO device descriptor
   tto_unit	TTO unit
   tto_reg	TTO register list
*/

UNIT tto_unit = { UDATA (&tto_svc, 0, 0), SERIAL_OUT_WAIT };

REG tto_reg[] = {
	{ ORDATA (BUF, tto_unit.buf, 6) },
	{ FLDATA (XFR, xfr_req, XFR_V_TTO) },
	{ ORDATA (CHAN, tto_ch, 4) },
	{ DRDATA (POS, tto_unit.pos, 31), PV_LEFT },
	{ DRDATA (TIME, tto_unit.wait, 24), REG_NZ + PV_LEFT },
	{ ORDATA (ASCH, tto_ac, 4), REG_HRO },
	{ NULL }  };

MTAB tto_mod[] = {
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL", "CHANNEL",
		&set_chan, &show_chan, &tto_ac },
	{ 0 }  };

DEVICE tto_dev = {
	"TTO", &tto_unit, tto_reg, tto_mod,
	1, 10, 31, 1, 8, 8,
	NULL, NULL, &tto_reset,
	NULL, NULL, NULL };

/* Paper tape reader

   The paper tape readers is a streaming input device.  Once started, it
   continues to read until disconnected.  Leader before the current record
   is ignored; leader after the current record sets channel EndOfRecord.
*/

t_stat ptr_conn (uint32 inst, int32 mode)
{
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */

if (mode) CRETINS;					/* mode 1? err */
if (QINVC (ptr_ch, ptr_ac, new_ch)) CRETIOP;		/* inv conn? err */
ptr_ch = new_ch;					/* save chan */
ptr_sor = 1;						/* start of rec */
xfr_req = xfr_req & ~XFR_PTR;				/* clr xfr flag */
sim_activate (&ptr_unit, ptr_unit.wait);		/* activate */
return SCPE_OK;
}

t_stat ptr_disc (uint32 dev)
{
ptr_sor = 0;						/* clear state */
ptr_ch = 0;
xfr_req = xfr_req & ~XFR_PTR;				/* clr xfr flag */
sim_cancel (&ptr_unit);					/* deactivate unit */
return SCPE_OK;
}

t_stat ptr_read (uint32 dev, uint32 *dat)
{
xfr_req = xfr_req & ~XFR_PTR;				/* clr xfr flag */
if (ptr_unit.buf != odd_par[ptr_unit.buf & 077])	/* good parity? */
	chan_set_flag (ptr_ch - 1, CHF_ERR);		/* no, error */
*dat = ptr_unit.buf & 077;				/* get buf data */
return SCPE_OK;
}

/* Unit service */

t_stat ptr_svc (UNIT *uptr)
{
int32 temp = 0;

if ((ptr_unit.flags & UNIT_ATT) == 0) {			/* attached? */
	ptr_set_err ();					/* no, err, disc */
	CRETIOE (ptr_stopioe, SCPE_UNATT);  }
if ((temp = getc (ptr_unit.fileref)) == EOF) {		/* end of file? */
	ptr_set_err ();					/* yes, err, disc */
	if (feof (ptr_unit.fileref)) {			/* end of file? */
		if (ptr_stopioe) printf ("PTR end of file\n");
		else return SCPE_OK;  }
	else perror ("PTR I/O error");			/* I/O error */
	clearerr (ptr_unit.fileref);
	return SCPE_IOERR;  }
ptr_unit.pos = ptr_unit.pos + 1;			/* inc position */
if (temp) {						/* leader/gap? */
	ptr_unit.buf = temp & 0177;			/* no, save char */
	xfr_req = xfr_req | XFR_PTR;			/* set xfr flag */
	ptr_sor = 0;  }					/* in record */
else if (!ptr_sor)					/* end record? */
	 chan_set_flag (ptr_ch - 1, CHF_EOR);		/* ignore leader */
sim_activate (&ptr_unit, ptr_unit.wait);		/* get next char */
return SCPE_OK;
}

/* Fatal error */

void ptr_set_err (void)
{
chan_set_flag (ptr_ch - 1, CHF_EOR | CHF_ERR);		/* eor, error */
chan_disc (ptr_ch - 1);					/* disconnect */
xfr_req = xfr_req & ~XFR_PTR;				/* clear xfr */
sim_cancel (&ptr_unit);					/* stop */
ptr_ch = 0;
return;
}

/* Reset routine */

t_stat ptr_reset (DEVICE *dptr)
{
if (ptr_ch) chan_disc (ptr_ch - 1);			/* disconnect */
ptr_ch = 0;
ptr_sor = 0;						/* clear state */
ptr_unit.buf = 0;
xfr_req = xfr_req & ~XFR_PTR;				/* clr xfr flag */
sim_cancel (&ptr_unit);					/* deactivate unit */
return SCPE_OK;
}

/* Paper tape punch

   The paper tape punch is an asynchronous streaming output device.  That is,
   it can never cause a channel rate error; if no data is available, it waits. */

t_stat ptp_conn (uint32 inst, int32 mode)
{
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */

if (mode) CRETINS;					/* mode 1? err */
if (QINVC (ptp_ch, ptp_ac, new_ch)) CRETIOP;		/* inv conn? err */
ptp_ch = new_ch;					/* save chan */
ptp_ldr = (inst & CHC_NLDR)? 1: 0;			/* leader? */
xfr_req = xfr_req & ~XFR_PTP;				/* clr xfr flag */
sim_activate (&ptp_unit, ptp_unit.wait);		/* activate */
return SCPE_OK;
}

t_stat ptp_disc (uint32 dev)
{
ptp_ldr = 0;						/* clear state */
ptp_ch = 0;
xfr_req = xfr_req & ~XFR_PTP;				/* clr xfr flag */
sim_cancel (&ptp_unit);					/* deactivate unit */
return SCPE_OK;
}

t_stat ptp_write (uint32 dev, uint32 *dat)
{
xfr_req = xfr_req & ~XFR_PTP;				/* clr xfr flag */
sim_activate (&ptp_unit, ptp_unit.wait);		/* activate */
ptp_unit.buf = odd_par[(*dat) & 077];			/* save data */
return ptp_out (ptp_unit.buf);				/* punch w/ par */
}

/* Unit service */

t_stat ptp_svc (UNIT *uptr)
{
int32 i;
t_stat r = SCPE_OK;

if (ptp_ldr) {						/* need leader? */
	for (i = 0; i < 12; i++) {			/* punch leader */
		if (r = ptp_out (0)) break;  }
	ptp_ldr = 0;  }					/* clear flag */
chan_set_ordy (ptp_ch - 1);				/* ptp ready */
return r;
}

/* Punch I/O */

t_stat ptp_out (int32 dat)
{
if ((ptp_unit.flags & UNIT_ATT) == 0) {			/* attached? */
	ptp_set_err ();					/* no, disc, err */
	CRETIOE (ptp_stopioe, SCPE_UNATT);  }
if (putc (dat, ptp_unit.fileref) == EOF) {		/* I/O error? */
	ptp_set_err ();					/* yes, disc, err */
	perror ("PTP I/O error");			/* print msg */
	clearerr (ptp_unit.fileref);
	return SCPE_IOERR;  }
ptp_unit.pos = ptp_unit.pos + 1;			/* inc position */
return SCPE_OK;
}

/* Fatal error */

void ptp_set_err (void)
{
chan_set_flag (ptp_ch - 1, CHF_ERR);			/* error */
chan_disc (ptp_ch - 1);					/* disconnect */
xfr_req = xfr_req & ~XFR_PTP;				/* clear xfr */
sim_cancel (&ptp_unit);					/* stop */
ptp_ch = 0;
return;
}

/* Reset routine */

t_stat ptp_reset (DEVICE *dptr)
{
if (ptp_ch) chan_disc (ptp_ch - 1);			/* disconnect */
ptp_ch = 0;
ptp_ldr = 0;						/* clear state */
ptp_unit.buf = 0;
xfr_req = xfr_req & ~XFR_PTP;				/* clr xfr flag */
sim_cancel (&ptp_unit);					/* deactivate unit */
return SCPE_OK;
}

/* Typewriter input

   The typewriter input is an asynchronous input device.  That is, it can
   never cause a channel rate error; if no data is available, it waits. */

t_stat tti_conn (uint32 inst, int32 mode)
{
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */

if (mode) CRETINS;					/* mode 1? err */
if (QINVC (tti_ch, tti_ac, new_ch)) CRETIOP;		/* inv conn? err */
tti_ch = new_ch;					/* save chan */
xfr_req = xfr_req & ~XFR_TTI;				/* clr xfr flag */
return SCPE_OK;
}

t_stat tti_disc (uint32 dev)
{
tti_ch = 0;
xfr_req = xfr_req & ~XFR_TTI;				/* clr xfr flag */
return SCPE_OK;
}

t_stat tti_read (uint32 dev, uint32 *dat)
{
xfr_req = xfr_req & ~XFR_TTI;				/* clr xfr flag */
*dat = tti_unit.buf;					/* get buf data */
return SCPE_OK;
}

/* Unit service */

t_stat tti_svc (UNIT *uptr)
{
int32 temp;

sim_activate (&tti_unit, tti_unit.wait);		/* continue poll */
if ((temp = sim_poll_kbd ()) < SCPE_KFLAG) return temp;	/* no char or error? */
temp = temp & 0177;
tti_unit.pos = tti_unit.pos + 1;
tti_unit.buf = ascii_to_sds[temp];			/* internal rep */
if (tti_unit.buf) {					/* representable? */
	sim_putchar (temp);				/* echo */
	if (temp == 015) sim_putchar (012);		/* lf after cr */
	xfr_req = xfr_req | XFR_TTI;  }			/* set xfr flag */
else sim_putchar (007);					/* ding! */
return SCPE_OK;
}

t_stat tti_reset (DEVICE *dptr)
{
if (tti_ch) chan_disc (tti_ch - 1);			/* disconnect */
tti_ch = 0;
tti_unit.buf = 0;					/* clear state */
xfr_req = xfr_req & ~XFR_TTI;				/* clr xfr flag */
sim_activate (&tti_unit, tti_unit.wait);		/* start poll */
return SCPE_OK;
}

/* Typewriter output

   The typewriter output is an asynchronous streaming output device.  That is,
   it can never cause a channel rate error; if no data is available, it waits. */

t_stat tto_conn (uint32 inst, int32 mode)
{
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */

if (mode) CRETINS;					/* mode 1? err */
if (QINVC (tto_ch, tto_ac, new_ch)) CRETIOP;		/* inv conn? err */
tto_ch = new_ch;					/* save chan */
xfr_req = xfr_req & ~XFR_TTO;				/* clr xfr flag */
sim_activate (&tto_unit, tto_unit.wait);		/* activate */
return SCPE_OK;
}

t_stat tto_disc (uint32 dev)
{
tto_ch = 0;
xfr_req = xfr_req & ~XFR_TTO;				/* clr xfr flag */
sim_cancel (&tto_unit);					/* deactivate unit */
return SCPE_OK;
}

t_stat tto_write (uint32 dev, uint32 *dat)
{
int32 asc;
t_stat r = SCPE_OK;

xfr_req = xfr_req & ~XFR_TTO;				/* clr xfr flag */
tto_unit.buf = (*dat) & 077;				/* save data */
if (tto_unit.buf == TT_CR) asc = '\n';			/* control chars? */
else if (tto_unit.buf == TT_BS) asc = '\b';
else if (tto_unit.buf == TT_TB) asc = '\t';
else asc = sds_to_ascii[tto_unit.buf];			/* translate */
r = sim_putchar (asc);					/* output */
tto_unit.pos = tto_unit.pos + 1;			/* inc position */
sim_activate (&tto_unit, tto_unit.wait);		/* activate */
return r;
}

/* Unit service */

t_stat tto_svc (UNIT *uptr)
{
chan_set_ordy (tto_ch - 1);				/* tto rdy */
return SCPE_OK;
}

/* Reset routine */

t_stat tto_reset (DEVICE *dptr)
{
if (tto_ch) chan_disc (tto_ch - 1);			/* disconnect */
tto_ch = 0;
tto_unit.buf = 0;					/* clear state */
xfr_req = xfr_req & ~XFR_TTO;				/* clr xfr flag */
sim_cancel (&tto_unit);					/* deactivate unit */
return SCPE_OK;
}
